%Cascaded Tanks Calculation of the Bounds Using RK4 for fixed time step. 
%Fifty random draws (curves) are generated, with the draws the furthest 
%above and below the mean prediction thrown out. Next highest and 
%lowest draws at each point are used as the bounds.

%Load the data sets, set up vectors of inputs
S = load('Tank1.mat');
S2 = load('Tank2.mat');
u1 = S.u';
y1 = S.y';
u2 = S2.u';
y2 = S2.y';

h1_T1 = y1(:,1);
h1_T2 = y2(:,1);
h2_T1 = y1(:,2);
h2_T2 = y2(:,2);

%Load the derivative models created in CasTanks_GP_5fold, and the spline coefficients
load('GP 5Fold Models.mat')
x = dlmread('spline_coefficient_500.txt');
phis = splineconvert500(x);

%Generate 50 random integers between 1 and the number of draws to use in
%selecting which draws to calculate
a1 = randperm(size(betash1{1},1),50);
a2 = randperm(size(betash2{1},1),50);

%Initialize cells to store the bounds for each fold
h1max = cell(5,1);
h1min = cell(5,1);
h2max = cell(5,1);
h2min = cell(5,1);

%Divide the data up into the same 5 folds as in CasTanks_GP_5fold. Also
%specify the timespan to integrate over for each fold. The 2nd fold
%contains two curves, so it must have two seperate calls to the integrator
%in order to calculate the bounds for each curve.
for i=1:5
    if i==1
        start=1*5; %Time at which integration begins
        stop=2000*5; %Time at which integration ends
        tspan=1:2000;
        tspan=tspan*5;
        y0=y1(1,:)'; %Initial condition
        test_h1 = h1_T1(1:2000);
        test_h2 = h2_T1(1:2000);
        h=0.5; %Time Step
    else
        if i==2
            start=2001*5; %Time at which integration begins for the 1st curve in this fold
            stop=2500*5; %Time at which integration ends for the 1st curve in this fold
            tspan=2001:2500;
            tspan=tspan*5;
            y0=y1(2001,:)'; %Initial condition for the 1st curve in this fold
            start2=1*4; %Time at which integration begins for the 2nd curve in this fold
            stop2=1500*4; %Time at which integration ends for the 2nd curve in this fold
            tspan2=1:1500;
            tspan2=tspan2*4;
            y0_2=y2(1,:)'; %Initial condition for the 2nd curve in this fold
            test_h1 = cell(2,1);
            test_h2 = cell(2,1);
            test_h1{1} = h1_T1(2001:end);
            test_h1{2} = h1_T2(1:1500);
            test_h2{1} = h2_T1(2001:end);
            test_h2{2} = h2_T2(1:1500);
            h=0.5; %Time Step for 1st curve in this fold
            h2=0.4; %Time Step for 2nd curve in this fold
        else
            if i==3
                start=1501*4; %Time at which integration begins
                stop=3500*4; %Time at which integration ends
                tspan=1501:3500;
                tspan=tspan*4;
                y0=y2(1501,:)'; %Initial condition
                test_h1 = h1_T2(1501:3500);
                test_h2 = h2_T2(1501:3500);
                h=0.4; %Time Step
            else
                if i==4
                    start=3501*4; %Time at which integration begins
                    stop=5500*4; %Time at which integration ends
                    tspan=3501:5500;
                    tspan=tspan*4;
                    y0=y2(3501,:)'; %Initial condition
                    test_h1 = h1_T2(3501:5500);
                    test_h2 = h2_T2(3501:5500);
                    h=0.4; %Time Step
                else
                    start=5501*4; %Time at which integration begins
                    stop=7500*4; %Time at which integration ends
                    tspan=5501:7500;
                    tspan=tspan*4;
                    y0=y2(5501,:)'; %Initial condition
                    test_h1 = h1_T2(5501:end);
                    test_h2 = h2_T2(5501:end);
                    h=0.4; %Time Step
                end
            end
        end
    end

    %Calculate the mean model
    mean_betash1 = mean(betash1{i});
    mean_betash2 = mean(betash2{i});
    meanbetash = {mean_betash1;mean_betash2};
    matrixh = {matrixh1{i};matrixh2{i}};
    if i~=2
        if i==1
            [~,mh1] = rk4_mean(meanbetash,matrixh,u1,norms{i},phis,start,stop,y0,h,1);
        else
            [~,mh1] = rk4_mean(meanbetash,matrixh,u2,norms{i},phis,start,stop,y0,h,2);
        end
    else
        [~,mh1] = rk4_mean(meanbetash,matrixh,u1,norms{i},phis,start,stop,y0,h,1);
        [~,mh2] = rk4_mean(meanbetash,matrixh,u2,norms{i},phis,start2,stop2,y0_2,h2,2);
    end

    ha = cell(50,5); %ha is a cell matrix of all the draws (for h1 and h2) evaluated.
    %1st dimension of ha represents which draw, 2nd dimension of ha
    %indicates which fold
    for k=1:50
        betash = {betash1{i};betash2{i}};
        if i~=2
            if i==1
                [Th1,ha{k,i}] = rk4(betash,matrixh,u1,norms{i},phis,start,stop,y0,h,1,a1(k),a2(k));
            else
                [Th1,ha{k,i}] = rk4(betash,matrixh,u2,norms{i},phis,start,stop,y0,h,2,a1(k),a2(k));
            end
        else
            %For the 2nd fold we have two curves in the fold, so the process needs
            %repeated for each curve. This 2nd curve's results called hb

            [Th1,ha{k,i}] = rk4(betash,matrixh,u1,norms{i},phis,start,stop,y0,h,1,a1(k),a2(k));
            [Th2,hb{k}] = rk4(betash,matrixh,u2,norms{i},phis,start2,stop2,y0_2,h2,2,a1(k),a2(k));

        end
    end

    
    %Determine the curves the furthest away from the mean prediction (one above
    %and one below) and discard them
    above_1=0;
    below_1=0;
    above_2=0;
    below_2=0;
    most_above_1 = 0;
    most_below_1 = 0;
    most_above_2 = 0;
    most_below_2 = 0;
    for k=1:50
        for l=1:length(ha{k,i})
            if (ha{k,i}(1,l)-mh1(1,l)) > above_1
                above_1 = (ha{k,i}(1,l)-mh1(1,l));
                most_above_1 = k;
            else
                if (ha{k,i}(1,l)-mh1(1,l)) < below_1
                    below_1 = (ha{k,i}(1,l)-mh1(1,l));
                    most_below_1 = k;
                end
            end
            if (ha{k,i}(2,l)-mh1(2,l)) > above_2
                above_2 = (ha{k,i}(2,l)-mh1(2,l));
                most_above_2 = k;
            else
                if (ha{k,i}(2,l)-mh1(2,l)) < below_2
                    below_2 = (ha{k,i}(2,l)-mh1(2,l));
                    most_below_2 = k;
                end
            end
        end
    end

    h1b = zeros(length(ha)-2,length(ha{1,i}));
    h2b = zeros(length(ha)-2,length(ha{1,i}));
    ind1 = 1;
    ind2 = 1;
    for ii=1:length(ha)
        if ii~= most_above_1 && ii~= most_below_1
            h1b(ind1,:) = ha{ii,i}(1,:);
            ind1 = ind1 + 1;
        end
        if ii~= most_above_2 && ii~= most_below_2
            h2b(ind2,:) = ha{ii,i}(2,:);
            ind2 = ind2+1;
        end
    end
    if i==2
        %Repeat the process of determining the bounds for the 2nd curve in
        %the 2nd fold.
        above_1=0;
        below_1=0;
        above_2=0;
        below_2=0;
        most_above_1 = 0;
        most_below_1 = 0;
        most_above_2 = 0;
        most_below_2 = 0;
        for k=1:50
            for l=1:length(hb{k})
                if (hb{k}(1,l)-mh2(1,l)) > above_1
                    above_1 = (hb{k}(1,l)-mh2(1,l));
                    most_above_1 = k;
                else
                    if (hb{k}(1,l)-mh2(1,l)) < below_1
                        below_1 = (hb{k}(1,l)-mh2(1,l));
                        most_below_1 = k;
                    end
                end
                if (hb{k}(2,l)-mh2(2,l)) > above_2
                    above_2 = (hb{k}(2,l)-mh2(2,l));
                    most_above_2 = k;
                else
                    if (hb{k}(2,l)-mh2(2,l)) < below_2
                        below_2 = (hb{k}(2,l)-mh2(2,l));
                        most_below_2 = k;
                    end
                end
            end
        end


        ind1 = 1;
        ind2 = 1;
        for ii=1:length(hb)
            if ii~= most_above_1 && ii~= most_below_1
                h1b_b(ind1,:) = hb{ii}(1,:);
                ind1 = ind1 + 1;
            end
            if ii~= most_above_2 && ii~= most_below_2
                h2b_b(ind2,:) = hb{ii}(2,:);
                ind2 = ind2+1;
            end
        end

        %Concatenate the two curves in the 2nd folds bounds together
        h1b = [h1b,h1b_b];
        h2b = [h2b,h2b_b];
    end

    %Set the bounds equal to the highest and lowest remaining
    %predictions at each point
    h1max{i} = max(h1b);
    h1min{i} = min(h1b);
    h2max{i} = max(h2b);
    h2min{i} = min(h2b);

    %Plot the bounds against the mean prediction
    if i~=2
        figure
        plot(Th1,mh1(1,:),'b')
        hold on
        plot(Th1,h1max{i},'r')
        hold on
        plot(Th1,h1min{i},'r')
        legend('Mean Prediction','Bounds')
        xlabel('Time')
        ylabel('Height')
        hold off

        figure
        plot(Th1,mh1(2,:),'b')
        hold on
        plot(Th1,h2max{i},'r')
        hold on
        plot(Th1,h2min{i},'r')
        legend('Mean Prediction','Bounds')
        xlabel('Time')
        ylabel('Height')
        hold off
    else
        figure
        plot(Th1,mh1(1,:),'b')
        hold on
        plot(Th1,h1max{i}(1:length(Th1)),'r')
        hold on
        plot(Th1,h1min{i}(1:length(Th1)),'r')
        legend('Mean Prediction','Bounds')
        xlabel('Time')
        ylabel('Height')
        hold off

        figure
        plot(Th1,mh1(2,:),'b')
        hold on
        plot(Th1,h2max{i}(1:length(Th1)),'r')
        hold on
        plot(Th1,h2min{i}(1:length(Th1)),'r')
        legend('Mean Prediction','Bounds')
        xlabel('Time')
        ylabel('Height')
        hold off

        figure
        plot(Th2,mh2(1,:),'b')
        hold on
        plot(Th2,h1max{i}(1+length(Th1):end),'r')
        hold on
        plot(Th2,h1min{i}(1+length(Th1):end),'r')
        legend('Mean Prediction','Bounds')
        xlabel('Time')
        ylabel('Height')
        hold off

        figure
        plot(Th2,mh2(2,:),'b')
        hold on
        plot(Th2,h2max{i}(1+length(Th1):end),'r')
        hold on
        plot(Th2,h2min{i}(1+length(Th1):end),'r')
        legend('Mean Prediction','Bounds')
        xlabel('Time')
        ylabel('Height')
        hold off
    end
end







function [T,Y]=rk4(betash,matrixh,u,norms,phis,start,stop,y0,h,ds,a1,a2)%Inputs are h1,h2,u; 
%ds indicates whether we are evaluating from the 1st data set or the 2nd
%(which affects the sampling frequency of u)
if ds==1
f=@(t,y) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/5)),norms(1,1),norms(2,1))], betash{1}(a1,:), phis, matrixh{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/5)),norms(1,1),norms(2,1))], betash{2}(a2,:), phis, matrixh{2})];
else
f=@(t,y) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/4)),norms(1,1),norms(2,1))], betash{1}(a1,:), phis, matrixh{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/4)),norms(1,1),norms(2,1))], betash{2}(a2,:), phis, matrixh{2})];
end
T=start:h:stop;
y=y0;
Y=[y0,zeros(2,length(T)-1)];
ind = 2;
for t=T(1:end-1)
    dy1=f(t,y)*h; %euler estimate of y(t+h)-y(t)
    %In between each estimate of the derivative we check to see if the
    %value of the outputs exceeds the range of the training data, and if
    %the derivative would produce a new prediction further outside the
    %accepted range. If both conditions are true, the derivative is set
    %equal to 0.
    if y(1)>norms(2,2) && dy1(1)>0
        dy1(1)=0;
    else
        if y(1)<norms(1,2) && dy1(1)<0
            dy1(1)=0;
        end
    end
    if y(2)>norms(2,3) && dy1(2)>0
        dy1(2)=0;
    else
        if y(2)<norms(1,3) && dy1(2)<0
            dy1(2)=0;
        end
    end
    dy2=f(t+h/2,y+dy1/2)*h; %midpoint estimate
    if (y(1)+dy1(1)/2)>norms(2,2) && dy2(1)>0
        dy2(1)=0;
    else
       if (y(1)+dy1(1)/2)<norms(1,2) && dy2(1)<0
            dy2(1)=0;
        end
    end
    if (y(2)+dy1(2)/2)>norms(2,3) && dy2(2)>0
        dy2(2)=0;
    else
        if (y(2)+dy1(2)/2)<norms(1,3) && dy2(2)<0
            dy2(2)=0;
        end
    end
    dy3=f(t+h/2,y+dy2/2)*h; %refined midpoint method
    if (y(1)+dy2(1)/2)>norms(2,2) && dy3(1)>0
        dy3(1)=0;
    else
        if (y(1)+dy2(1)/2)<norms(1,2) && dy3(1)<0
            dy3(1)=0;
        end
    end
    if (y(2)+dy2(2)/2)>norms(2,3) && dy3(2)>0
        dy3(2)=0;
    else
        if (y(2)+dy2(2)/2)<norms(1,3) && dy3(2)<0
            dy3(2)=0;
        end
    end
    dy4=f(t+h,y+dy3)*h; %forward estimate
    if (y(1)+dy3(1))>norms(2,2) && dy4(1)>0
        dy4(1)=0;
    else
        if (y(1)+dy3(1))<norms(1,2) && dy4(1)<0
            dy4(1)=0;
        end
    end
    if (y(2)+dy3(2))>norms(2,3) && dy4(2)>0
        dy4(2)=0;
    else
        if (y(2)+dy3(2))<norms(1,3) && dy4(2)<0
            dy4(2)=0;
        end
    end
    y=y+(dy1+2*dy2+2*dy3+dy4)/6; %averaging
    Y(:,ind)=y; %augment with a new column
    ind = ind+1;
end
end

function [T,Y]=rk4_mean(meanbetash,matrixh,u,norms,phis,start,stop,y0,h,ds)%Inputs is h1,h2,u,h1l,h2l,ul; ds indicates whether we are evaluating ds 1 or 2
if ds==1
f=@(t,y) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/5)),norms(1,1),norms(2,1))], meanbetash{1}, phis, matrixh{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/5)),norms(1,1),norms(2,1))], meanbetash{2}, phis, matrixh{2})];
else
f=@(t,y) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/4)),norms(1,1),norms(2,1))], meanbetash{1}, phis, matrixh{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/4)),norms(1,1),norms(2,1))], meanbetash{2}, phis, matrixh{2})];
end
T=start:h:stop;
y=y0;
Y=[y0,zeros(2,length(T)-1)];
ind = 2;
for t=T(1:end-1)
    %dy = f(t,y)
    %inp_2 = [norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/4)),norms(1,1),norms(2,1))]
    dy1=f(t,y)*h; %euler estimate of y(t+h)-y(t)
    if y(1)>norms(2,2) && dy1(1)>0
        dy1(1)=0;
    else
        if y(1)<norms(1,2) && dy1(1)<0
            dy1(1)=0;
        end
    end
    if y(2)>norms(2,3) && dy1(2)>0
        dy1(2)=0;
    else
        if y(2)<norms(1,3) && dy1(2)<0
            dy1(2)=0;
        end
    end
    dy2=f(t+h/2,y+dy1/2)*h; %midpoint estimate
    if (y(1)+dy1(1)/2)>norms(2,2) && dy2(1)>0
        dy2(1)=0;
    else
       if (y(1)+dy1(1)/2)<norms(1,2) && dy2(1)<0
            dy2(1)=0;
        end
    end
    if (y(2)+dy1(2)/2)>norms(2,3) && dy2(2)>0
        dy2(2)=0;
    else
        if (y(2)+dy1(2)/2)<norms(1,3) && dy2(2)<0
            dy2(2)=0;
        end
    end
    dy3=f(t+h/2,y+dy2/2)*h; %refined midpoint method
    if (y(1)+dy2(1)/2)>norms(2,2) && dy3(1)>0
        dy3(1)=0;
    else
        if (y(1)+dy2(1)/2)<norms(1,2) && dy3(1)<0
            dy3(1)=0;
        end
    end
    if (y(2)+dy2(2)/2)>norms(2,3) && dy3(2)>0
        dy3(2)=0;
    else
        if (y(2)+dy2(2)/2)<norms(1,3) && dy3(2)<0
            dy3(2)=0;
        end
    end
    dy4=f(t+h,y+dy3)*h; %forward estimate
    if (y(1)+dy3(1))>norms(2,2) && dy4(1)>0
        dy4(1)=0;
    else
        if (y(1)+dy3(1))<norms(1,2) && dy4(1)<0
            dy4(1)=0;
        end
    end
    if (y(2)+dy3(2))>norms(2,3) && dy4(2)>0
        dy4(2)=0;
    else
        if (y(2)+dy3(2))<norms(1,3) && dy4(2)<0
            dy4(2)=0;
        end
    end
    y=y+(dy1+2*dy2+2*dy3+dy4)/6; %averaging
    Y(:,ind)=y; %augment with a new column
    ind = ind+1;
end
end

function ny = norma(y,n1,n2) %0 to 1 Normalization subfunction.
%y is the actual value, n1 is the minimum, n2 is the maximum
if y > n2
    ny= 1;
else
    if y < n1
        ny = 0;
    else
        ny = (y-n1)/(n2-n1);
    end
end
end