%Cascaded Tanks Integration No Lag Using RK4 for fixed time step


%Load the data sets, set up vectors of inputs
S = load('Tank1.mat');
S2 = load('Tank2.mat');
u1 = S.u';
y1 = S.y';
u2 = S2.u';
y2 = S2.y';

h1_T1 = y1(:,1);
h1_T2 = y2(:,1);
h2_T1 = y1(:,2);
h2_T2 = y2(:,2);

%Load the derivative models created in CasTanks_GP_5fold
load('GP 5Fold Models.mat')
%Load the spline coefficients
x = dlmread('spline_coefficient_500.txt');
phis = splineconvert500(x);

%Initialize measurement values
MAE_h1 = zeros(5,1);
MAE_h2 = zeros(5,1);
MAPE_h1 = zeros(5,1);
MAPE_h2 = zeros(5,1);

T_Predict = zeros(5,1);

%Divide the data up into the same 5 folds as in CasTanks_GP_5fold. Also
%specify the timespan to integrate over for each fold. The 2nd fold
%contains two curves, so it must have two seperate calls to the integrator.
for i=1:5
    if i==1
        start=1*5; %Time at which integration begins
        stop=2000*5;%Time at which integration ends
        tspan=1:2000;
        tspan=tspan*5;
        y0=y1(1,:)'; %Initial condition
        test_h1 = h1_T1(1:2000);
        test_h2 = h2_T1(1:2000);
        h=0.5; %Time Step
    else
        if i==2
            start=2001*5; %Time at which integration begins for 1st curve in the fold
            stop=2500*5; %Time at which integration ends for 2nd curve in this fold
            tspan=2001:2500;
            tspan=tspan*5;
            y0=y1(2001,:)'; %Initial condition for 1st curve in this fold
            start2=1*4; %Time at which integration begins for 2nd curve in this fold
            stop2=1500*4; %Time at which integration ends for 2nd curve in this fold
            tspan2=1:1500;
            tspan2=tspan2*4;
            y0_2=y2(1,:)'; %Initial condition for 2nd curve in this fold
            test_h1 = cell(2,1);
            test_h2 = cell(2,1);
            test_h1{1} = h1_T1(2001:end);
            test_h1{2} = h1_T2(1:1500);
            test_h2{1} = h2_T1(2001:end);
            test_h2{2} = h2_T2(1:1500);
            h=0.5; %Time Step for Integration of 1st curve in this fold
            h2=0.4; %Time Step for Integration of 2nd curve in this fold
        else
            if i==3
                start=1501*4; %Time at which integration begins
                stop=3500*4; %Time at which integration ends
                tspan=1501:3500;
                tspan=tspan*4;
                y0=y2(1501,:)'; %Initial condition
                test_h1 = h1_T2(1501:3500);
                test_h2 = h2_T2(1501:3500);
                h=0.4; %Time Step
            else
                if i==4
                    start=3501*4; %Time at which integration begins
                    stop=5500*4; %Time at which integration ends
                    tspan=3501:5500;
                    tspan=tspan*4;
                    y0=y2(3501,:)'; %Initial condition
                    test_h1 = h1_T2(3501:5500);
                    test_h2 = h2_T2(3501:5500);
                    h=0.4; %Time Step
                else
                    start=5501*4; %Time at which integration begins
                    stop=7500*4; %Time at which integration ends
                    tspan=5501:7500;
                    tspan=tspan*4;
                    y0=y2(5501,:)'; %Initial condition
                    test_h1 = h1_T2(5501:end);
                    test_h2 = h2_T2(5501:end);
                    h=0.4; %Time Step
                end
            end
        end
    end

    %Calculate the mean GP model to make predictions with
    mean_betash1 = mean(betash1{i});
    mean_betash2 = mean(betash2{i});

    %Combine the models for dh1/dt and dh2/dt into a pair of inputs for
    %the RK4 integrator
    mean_betash = {mean_betash1;mean_betash2};
    matrixh = {matrixh1{i};matrixh2{i}};

    %For all folds but the 2nd we only have one curve per fold and use the
    %following setup
    if i~=2
        %Perform the integration of the model (and measure the time it takes)
        if i==1
            T_PredictS = tic;
            [Th1,h1] = rk4(mean_betash,matrixh,u1,norms{i},phis,start,stop,y0,h,1);
            T_Predict(i) = toc(T_PredictS);
        else
            T_PredictS = tic;
            [Th1,h1] = rk4(mean_betash,matrixh,u2,norms{i},phis,start,stop,y0,h,2);
            T_Predict(i) = toc(T_PredictS);
        end


        %Plot the predicted curve against the actual test curve
        figure
        plot(Th1,h1(1,:) ,'b')
        hold on
        plot(tspan,test_h1,'r')
        hold off
        figure
        plot(Th1,h1(2,:),'b')
        hold on
        plot(tspan,test_h2,'r')
        hold off

        %Reduce the number of datapoints from the integrated model so as to be
        %comparible to the true values
        y_int_h1 = h1(1,1:10:end);
        y_int_h2 = h1(2,1:10:end);

        %Calculate the MAE and MAPE for this particular fold
        MAE_h1(i) = mean(abs(y_int_h1'-test_h1));
        MAPE_h1(i) = mean(abs(((y_int_h1')-test_h1)./test_h1));
        MAE_h2(i) = mean(abs(y_int_h2'-test_h2));
        MAPE_h2(i) = mean(abs(((y_int_h2')-test_h2)./test_h2));

    else
        %For the 2nd fold since we have two curves we need to integrate twice and
        %use the weighted average of the MAE and MAPE (vs the number of datapoints
        %in the two curves) for their respective values for this fold.

        %Perform the two integrations. The amount of time measured is equal
        %to the time it takes to integrate both models
        T_PredictS = tic;
        [Th1,h1] = rk4(mean_betash,matrixh,u1,norms{i},phis,start,stop,y0,h,1);
        [Th2,h2] = rk4(mean_betash,matrixh,u2,norms{i},phis,start2,stop2,y0_2,h2,2);
        T_Predict(i) = toc(T_PredictS);

        %Plot both curves vs the actual results
        figure
        plot(Th1,h1(1,:) ,'b')
        hold on
        plot(tspan,test_h1{1},'r')
        hold off
        figure
        plot(Th1,h1(2,:),'b')
        hold on
        plot(tspan,test_h2{1},'r')
        hold off

        figure
        plot(Th2,h2(1,:) ,'b')
        hold on
        plot(tspan2,test_h1{2},'r')
        hold off
        figure
        plot(Th2,h2(2,:),'b')
        hold on
        plot(tspan2,test_h2{2},'r')
        hold off

        %Reduce the number of datapoints from the integrated models so as to be
        %comparible to the true values
        y_int_h1_1 = h1(1,1:10:end);
        y_int_h2_1 = h1(2,1:10:end);
        y_int_h1_2 = h2(1,1:10:end);
        y_int_h2_2 = h2(2,1:10:end);

        %Calculate the weighted average of the MAE and MAPE for this fold
        MAE_h1_1 = mean(abs(y_int_h1_1'-test_h1{1}));
        MAPE_h1_1 = mean(abs(((y_int_h1_1')-test_h1{1})./test_h1{1}));
        MAE_h2_1 = mean(abs(y_int_h2_1'-test_h2{1}));
        MAPE_h2_1 = mean(abs(((y_int_h2_1')-test_h2{1})./test_h2{1}));
        MAE_h1_2 = mean(abs(y_int_h1_2'-test_h1{2}));
        MAPE_h1_2 = mean(abs(((y_int_h1_2')-test_h1{2})./test_h1{2}));
        MAE_h2_2 = mean(abs(y_int_h2_2'-test_h2{2}));
        MAPE_h2_2 = mean(abs(((y_int_h2_2')-test_h2{2})./test_h2{2}));

        MAE_h1(i) = MAE_h1_1*0.25 + MAE_h1_2*0.75;
        MAE_h2(i) = MAE_h2_1*0.25 + MAE_h2_2*0.75;
        MAPE_h1(i) = MAPE_h1_1*0.25 + MAPE_h1_2*0.75;
        MAPE_h2(i) = MAPE_h2_1*0.25 + MAPE_h2_2*0.75;
    end
end

%Calculate the average MAE and MAPE values over all 5 folds
MAE_h1_tot = mean(MAE_h1)
MAE_h2_tot = mean(MAE_h2)
MAPE_h1_tot = mean(MAPE_h1)
MAPE_h2_tot = mean(MAPE_h2)

%Calculate the standard deviation of the MAE and MAPE values between folds
MAE_h1_stdev = (sum((MAE_h1-MAE_h1_tot).^2)/5)^0.5
MAE_h2_stdev = (sum((MAE_h2-MAE_h2_tot).^2)/5)^0.5
MAPE_h1_stdev = (sum((MAPE_h1-MAPE_h1_tot).^2)/5)^0.5
MAPE_h2_stdev = (sum((MAPE_h2-MAPE_h2_tot).^2)/5)^0.5

%Calculate the mean time it takes to integrate the model and produce a
%prediction for the curve.
Mean_Predict_Time = mean(T_Predict)


function [T,Y]=rk4(mean_betash,matrixh,u,norms,phis,start,stop,y0,h,ds)%Inputs are h1,h2,u; 
%ds indicates whether we are evaluating a curve from the 1st or 2nd
%dataset. As the two datasets have different sampling frequencies, this
%affects the equation for the current value of u.
if ds==1
f=@(t,y) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/5)),norms(1,1),norms(2,1))], mean_betash{1}, phis, matrixh{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/5)),norms(1,1),norms(2,1))], mean_betash{2}, phis, matrixh{2})];
else
f=@(t,y) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/4)),norms(1,1),norms(2,1))], mean_betash{1}, phis, matrixh{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(u(floor(t/4)),norms(1,1),norms(2,1))], mean_betash{2}, phis, matrixh{2})];
end
T=start:h:stop;
y=y0;
Y=[y0,zeros(2,length(T)-1)];
ind = 2;
for t=T(1:end-1)
    dy1=f(t,y)*h; %euler estimate of y(t+h)-y(t)
    %In between each estimate of the derivative we check to see if the
    %value of the outputs exceeds the range of the training data, and if
    %the derivative would produce a new prediction further outside the
    %accepted range. If both conditions are true, the derivative is set
    %equal to 0.
    if y(1)>norms(2,2) && dy1(1)>0
        dy1(1)=0;
    else
        if y(1)<norms(1,2) && dy1(1)<0
            dy1(1)=0;
        end
    end
    if y(2)>norms(2,3) && dy1(2)>0
        dy1(2)=0;
    else
        if y(2)<norms(1,3) && dy1(2)<0
            dy1(2)=0;
        end
    end
    dy2=f(t+h/2,y+dy1/2)*h; %midpoint estimate
    if (y(1)+dy1(1)/2)>norms(2,2) && dy2(1)>0
        dy2(1)=0;
    else
       if (y(1)+dy1(1)/2)<norms(1,2) && dy2(1)<0
            dy2(1)=0;
        end
    end
    if (y(2)+dy1(2)/2)>norms(2,3) && dy2(2)>0
        dy2(2)=0;
    else
        if (y(2)+dy1(2)/2)<norms(1,3) && dy2(2)<0
            dy2(2)=0;
        end
    end
    dy3=f(t+h/2,y+dy2/2)*h; %refined midpoint method
    if (y(1)+dy2(1)/2)>norms(2,2) && dy3(1)>0
        dy3(1)=0;
    else
        if (y(1)+dy2(1)/2)<norms(1,2) && dy3(1)<0
            dy3(1)=0;
        end
    end
    if (y(2)+dy2(2)/2)>norms(2,3) && dy3(2)>0
        dy3(2)=0;
    else
        if (y(2)+dy2(2)/2)<norms(1,3) && dy3(2)<0
            dy3(2)=0;
        end
    end
    dy4=f(t+h,y+dy3)*h; %forward estimate
    if (y(1)+dy3(1))>norms(2,2) && dy4(1)>0
        dy4(1)=0;
    else
        if (y(1)+dy3(1))<norms(1,2) && dy4(1)<0
            dy4(1)=0;
        end
    end
    if (y(2)+dy3(2))>norms(2,3) && dy4(2)>0
        dy4(2)=0;
    else
        if (y(2)+dy3(2))<norms(1,3) && dy4(2)<0
            dy4(2)=0;
        end
    end
    y=y+(dy1+2*dy2+2*dy3+dy4)/6; %averaging
    Y(:,ind)=y; %augment with a new column
    ind = ind+1;
end
end

function ny = norma(y,n1,n2) %0 to 1 Normalization subfunction.
%y is the actual value, n1 is the minimum, n2 is the maximum
if y > n2
    ny= 1;
else
    if y < n1
        ny = 0;
    else
        ny = (y-n1)/(n2-n1);
    end
end
end