function [vderiv, sm_v] = Derivative_GP_SIR(t, v, draws,obsError,a,b,atau,btau,tolerance,aic)
%Function that calculate the derivative of a vector, v, wrt another vector
%t using a Gaussian Process. A model of v with respect to t is created, then that model
%is differentiated. This takes advantage of the smoothing that is naturally
%present in the Gaussian Process. Higher order derivatives could be
%calculated in the same manner but are not currently required.

dpt = length(t);
dp = length(v);


%Check to confirm t and v are the same length
if(dpt ~= dp)
    ME = MException('1stGP:wrongsize', 'length of input must be the same as the length of the data');
    throw(ME)
end

%Normalize our only input
NormTime = normalize(t);

%Load the spline coefficients
x = dlmread('spline_coefficient_500.txt');
phis = splineconvert500(x);

%Create a GP model
[beta,matrix,~] = emulator_Xin(obsError, NormTime, v, 1, phis, a, b, atau, btau, tolerance, draws, 0, 0, aic);

%Only use the last half of the draws we produced
start = ceil(size(beta,1)/2);
beta = beta(start:end,:);
tol = 0;

%Use the mean prediction to produce a smoothed version of the curve:
mbeta = mean(beta);
sm_v = bss_eval(NormTime, mbeta, phis, matrix);

draws = size(beta,1);
used_Draws = randi(draws,dp,1); %Select a random assortment of the remaining draws to use when evaluating the derivative

deriv = 1; %Only the 1st order derivative is calculated
vderiv = zeros(dp,1);
%Calculate the derivatives
for i=1:dp
    vderiv(i) = bss_derivatives(NormTime(i), beta(used_Draws(i),:), phis, matrix, deriv, range(t));   
end

end