function [curves,inputs, pureinputs]=SIR_GP_Data_Test(p)
%if p = 1 will plot the curves
%y(,1) = S, y(,2) = I, y(,3) = R
%Error in y(2) = I is 1/4 that in S,R
%Generates virtual data for SIR model; for b values that vary based off of
%a sinusoidal or ramp forcing function and different ICs (starting values of S, I, and R)
%All curves should have the same number of data points (350)

%Total number of people (y(1) +y(2) +y(3)) equals 1000
b = [1.35:3.4:8.15]; %Starting b values
g = 0.5; %Held constant for everything

curves = cell(length(b),1);
inputs = cell(length(b),1);
pureinputs = cell(length(b),1);
tspan0 = [0:0.1:35]; %The initial timespan

y0 = cell(length(b),1);

%Our 12 inital conditions
y0{1} = [599,774,645,520;400,225,5,5;1,1,350,475];
y0{2} = [994,895,599,570;5,5,400,5;1,100,1,425];
y0{3} = [994,920,499,470;5,5,600,5;1,75,1,525];
ind = 1;

t_1=0; %Start all FFs at t=0
t_2=4; %End all ramps at t=4;
slope = 1; %So b will rise up to b_i+4
amp1 = 0.5; %For low and high b_i values use relatively small amplitude
amp2 = 3; %Use much higher amplitude for b_i in the center of the input space
per= 1;%Use the same period for all sinusoidal forcing functions


opts = odeset('MaxStep', 0.01);
for i = 1:length(b) 
    for l=1:size(y0{i},2)
        for j=1:length(g)
            for ii=1:2
                if ii==1
                    if i<3
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_B(t,y,b(i),g,slope,t_1,t_2), tspan0, y0{i}(:,l),opts); %Curves of postive ramp forcing function
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                if t(iii)<t_2
                                    B(iii) = b(i)+ slope*(t(iii)-t_1);
                                else
                                    B(iii) = b(i) + slope*(t_2-t_1);
                                end
                            end
                        end

                    else
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_B(t,y,b(i),g,(-1)*slope,t_1,t_2), tspan0, y0{i}(:,l),opts); %Curves of postive ramp forcing function.
                        % Give a downward slope to ramps in b that start high
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                if t(iii)<t_2
                                    B(iii) = b(i)- slope*(t(iii)-t_1);
                                else
                                    B(iii) = b(i) - slope*(t_2-t_1);
                                end
                            end
                        end
                    end
                else
                    if i~=2
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_sin(t,y,b(i),g,t_1,per,amp1), tspan0, y0{i}(:,l),opts); %Curves of low amplitude sinusoidal forcing function
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                B(iii) = b(i) + amp1*sin((t(iii)-t_1)*2*pi()/per);
                            end
                        end
                    else
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_sin(t,y,b(i),g,t_1,per,amp2), tspan0, y0{i}(:,l),opts);%Curves of high amplitude sinusoidal forcing function
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                B(iii) = b(i) + amp2*sin((t(iii)-t_1)*2*pi()/per);
                            end
                        end
                    end
                end

            %Determine the point at which I drops below 2. Use this as the
            %new final point in the time span
            tspan1 = 0;
            for k=1:length(tspan0)
                if (y(k,2) < 2) && (length(tspan1) < 2)
                    tspan1 = [0:t(k)/350:t(k)];
                end
            end
            


            if length(tspan1) > 1

                if ii==1
                    if i<3
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_B(t,y,b(i),g,slope,t_1,t_2), tspan1, y0{i}(:,l),opts);%Recalculate curves with smaller timespan
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                if t(iii)<t_2
                                    B(iii) = b(i)+ slope*(t(iii)-t_1);
                                else
                                    B(iii) = b(i) + slope*(t_2-t_1);
                                end
                            end
                        end

                    else
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_B(t,y,b(i),g,(-1)*slope,t_1,t_2), tspan1, y0{i}(:,l),opts); %Recalculate curves with smaller timespan
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                if t(iii)<t_2
                                    B(iii) = b(i)- slope*(t(iii)-t_1);
                                else
                                    B(iii) = b(i) - slope*(t_2-t_1);
                                end
                            end
                        end
                    end
                else
                    if i~=2
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_sin(t,y,b(i),g,t_1,per,amp1), tspan1, y0{i}(:,l),opts);%Recalculate curves with smaller timespan
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                B(iii) = b(i) + amp1*sin((t(iii)-t_1)*2*pi()/per);
                            end
                        end
                    else
                        [t,y] = ode45(@(t,y) Toy_SIR_ODE_sin(t,y,b(i),g,t_1,per,amp2), tspan1, y0{i}(:,l),opts);%Recalculate curves with smaller timespan
                        B = zeros(size(t));
                        for iii=1:length(t)
                            if t(iii)<t_1
                                B(iii) = b(i);
                            else
                                B(iii) = b(i) + amp2*sin((t(iii)-t_1)*2*pi()/per);
                            end
                        end
                    end
                end

            end
            
            yp = y;
            %Error S,I,R:
            for ij = 1: length(t)
                %rand() produces random numbers from 0 to 1; subtracted by
                %0.5 to give half positive, half negative
                a = (rand()-0.5)*8;
                y(ij,1) = y(ij,1) + a;
                y(ij,2) = y(ij,2) + a/4;
                y(ij,3) = y(ij,3) + a;
            end
            curves{ind} = y;
            
            T = t;
            S = y(:,1);
            I = y(:,2);
            R = y(:,3);
            %Since gamma is held constant it is not included in the inputs
            inputs{ind} = [T,B,S,I,R];
            pureinputs{ind} = [T,B,yp];
            ind = ind+1;
            end
        end
    end
end

if p == 1 %Produces plots of input space and individual curves
    for j =1:(length(b)*length(g)*size(y0,2))
        figure
        for i=1:3
            plot(inputs{j}(:,1),curves{j}(:,i))
            hold on
        end
        hold off
    end
    figure
    for j=1:size(curves,1)
       plot(pureinputs{j}(:,5),pureinputs{j}(:,4))
       hold on
    end
    xlabel('R')
    ylabel('I')
    title('Input Space (Varying ICs, bs)')
    hold off
end
