function [curves,inputs, pureinputs]=SIR_GP_Data_Train(p)
%if p = 1 will plot the curves
%Generates virtual training data for SIR model; for fixed b values at
%different ICs (starting values of S, I, and R)
%y(,1) = S, y(,2) = I, y(,3) = R
%Error in y(2) = I is 1/4 that in S,R
%All curves should have the same number of data points (350)

%Total number of people (y(1) +y(2) +y(3)) equals 1000
b = [0.5:1.7:9];
g = 0.5; %Held constant for everything

curves = cell(length(b),1);
inputs = cell(length(b),1);
pureinputs = cell(length(b),1);
tspan0 = [0:0.1:35]; %Inital timespan to run the ode solver over

y0 = cell(length(b),1); %Our ICs:
y0{1} = [994,599,449,899,799,699,949,299;5,400,550,100,200,300,50,700;1,1,1,1,1,1,1,1];
y0{2} = [994,274,424,895,770,599,774,645,520,395;5,725,575,5,5,400,225,5,5,5;1,1,1,100,225,1,1,350,475,600];
y0{3} = [994,224,374,895,795,599,695,570,445,320;5,775,625,5,5,400,5,5,5,5;1,1,1,100,200,1,300,425,550,675];
y0{4} = [994,199,920,820,499,720,595,470,370,270;5,800,5,5,500,5,5,5,5,5;1,1,75,175,1,275,400,525,625,725];
y0{5} = [994,179,920,820,449,720,595,470,370,270;5,820,5,5,550,5,5,5,5,5;1,1,75,175,1,275,400,525,625,725];
y0{6} = [994,179,920,820,399,720,595,470,370,270;5,820,5,5,600,5,5,5,5,5;1,1,75,175,1,275,400,525,625,725];
ind = 1;

for i = 1:length(b)
    for l=1:size(y0{i},2)
        for j=1:length(g)
            
            [t,y] = ode45(@(t,y) Toy_SIR_ODE(t,y,b(i),g(j)), tspan0, y0{i}(:,l));
            %No error version S,I,R:
            tspan1 = 0;
            for k=1:length(tspan0)
                if (y(k,2) < 2) && (length(tspan1) < 2)
                    tspan1 = [0:t(k)/350:t(k)]; %Finds the time at which I falls below 2. Sets this equal to the new maximum time in the 2nd timespan
                end
            end
                
            if length(tspan1) > 1 %If I<2 over the 1st timespan, calculates the results for the shorter 2nd timespan

                [t,y] = ode45(@(t,y) Toy_SIR_ODE(t,y,b(i),g(j)), tspan1, y0{i}(:,l));
            end
            yp = y;
            %Noise added to S,I,R:
            for ii = 1: length(t)
                %rand() produces random numbers from 0 to 1; subtracted by
                %0.5 to give half positive, half negative
                a = (rand()-0.5)*8;
                y(ii,1) = y(ii,1) + a;
                y(ii,2) = y(ii,2) + a/4;
                y(ii,3) = y(ii,3) + a;
            end
            curves{ind} = y;
            B = zeros(length(t),1) + b(i);
            T = t;
            S = y(:,1);
            I = y(:,2);
            R = y(:,3);
            %Since gamma is held constant it is not included in the inputs
            inputs{ind} = [T,B,S,I,R];
            pureinputs{ind} = [T,B,yp];
            ind = ind+1;
        end
    end
end

if p == 1 %Produces plots of input space and individual curves
    for j =1:(length(b)*length(g)*size(y0,2))
        figure
        for i=1:3
            plot(inputs{j}(:,1),curves{j}(:,i))
            hold on
        end
        hold off
    end
    figure
    for j=1:size(curves,1)
       plot(pureinputs{j}(:,5),pureinputs{j}(:,4))
       hold on
    end
    xlabel('R')
    ylabel('I')
    title('Input Space (Varying ICs, bs)')
    hold off
end
