%Creates a model of the SIR derivatives using Gaussian Process. Only models for dI/dt,
%dR/dt created (inputs of each model are I,R,b and I respectively).
%The model produced in this code is saved and loaded in SIR_GP_RK4Int and
%SIR_GP_RK4Bounds where its predictions are integrated and evaluated

%Load the training data
[curves,inputs, pureinputs]=SIR_GP_Data_Train(0);


tot_numcurves = max(size(curves));

for i=1:tot_numcurves
curve_lengths(i) = size(curves{i},1);
end

%Hyperparameters for calculating the derivative
a = 4;
atau = 4;
firstDraws = 2000;

%Initialize cells to for the derivatives of I and R and smoothed I and R
%curves
derivI = cell(tot_numcurves,1);
derivR = cell(tot_numcurves,1);
sm_I = cell(tot_numcurves,1);
sm_R = cell(tot_numcurves,1);

%Differentiation process. Differentiation is carried out by making a GP
%model of one curve wrt time, then differentiating the spline eqns
for i=1:tot_numcurves
        h_I = curves{i}(:,2);
        h_R = curves{i}(:,3);
        h_t = inputs{i}(:,1);
        %Should be 4 for varS and varR, 0.25 for varI
        varR = 4;
        varI = 0.25;
        scaleI = max(h_I);
        scaleR = max(h_R);
        bI = varI*(a+1);
        bR = varR*(a+1);
        tauI = sqrt(scaleI/varI^2);
        tauR = sqrt(scaleR/varR^2);
        btauI = tauI*(atau+1);
        btauR = tauR*(atau+1);

        [derivI{i}, sm_I{i}] = Derivative_GP_SIR(h_t,h_I,firstDraws,varI,a,bI,atau,btauI,3,1);
        [derivR{i}, sm_R{i}] = Derivative_GP_SIR(h_t,h_R,firstDraws,varR,a,bR,atau,btauR,3,1);
        btausI(i) = btauI;
        btausR(i) = btauR;       
end

%Load the spline coefficients
x = dlmread('spline_coefficient_500.txt');
phis = splineconvert500(x);

%Concatenate all the curve's derivatives and inputs together. Smooth the
%inputs.
tot_derivI = [];
tot_derivS = [];
tot_derivR = [];
tot_pureinputs = [];
tot_smoothinputs = [];
for i =1:tot_numcurves
    tot_derivI = [tot_derivI; derivI{i}];
    tot_derivR = [tot_derivR; derivR{i}];
    tot_pureinputs = [tot_pureinputs;pureinputs{i}];
    tot_smoothinputs = [tot_smoothinputs; [sm_I{i},sm_R{i}]];
end
tot_smI = zeros(size(tot_derivI));
tot_smR = zeros(size(tot_derivI));
%If the smoothing process has caused any point to exceed the previous
%minimum or maximum value for that input, we set that point equal to the
%minimum or maximum respectively
for i=1:length(tot_derivI)
   if  tot_smoothinputs(i,1) > max(tot_pureinputs(:,4))
       tot_smI(i) = max(tot_pureinputs(:,4));
   else
       if tot_smoothinputs(i,1) < min(tot_pureinputs(:,4))
           tot_smI(i) = min(tot_pureinputs(:,4));
       else
           tot_smI(i) = tot_smoothinputs(i,1);
       end
   end
   if  tot_smoothinputs(i,2) > max(tot_pureinputs(:,5))
       tot_smR(i) = max(tot_pureinputs(:,5));
   else
       if tot_smoothinputs(i,2) < min(tot_pureinputs(:,5))
           tot_smR(i) = min(tot_pureinputs(:,5));
       else
           tot_smR(i) = tot_smoothinputs(i,2);
       end
   end
end

%Normalize the inputs
norm_b = normalize(tot_pureinputs(:,2));
norm_smI = normalize(tot_smI);
norm_smR = normalize(tot_smR);

%Set up matricies of the appropriate inputs for the two models
I_inputs = [norm_smI,norm_smR,norm_b];
R_inputs = norm_smI;

%Set up the remaining hyperparameters for the Gaussian Process
a2 = 4;
scale2I = abs(mean(tot_derivI));
scale2R = abs(mean(tot_derivR));
bI = varI*(a2+1);
bR = varR*(a2+1);
tau2I = sqrt(scale2I/varI^2);
tau2R = sqrt(scale2R/varR^2);
btau2I = tau2I*(atau+1);
btau2R = tau2R*(atau+1);
draws = 2000;
relatsI = [1;1;1;1;1;1];
relatsR = 1;

%Build the model on the training data, throw out the first 1000 draws, and
%measure the time it takes
T_BuildS = tic;
[over_betasI,matrixI,evsI] = emulator_Xin(varI, I_inputs, tot_derivI, relatsI, phis, a2, bI, atau, btau2I, 3, draws, 0, 1,0);
[over_betasR,matrixR,evsR] = emulator_Xin(varR, R_inputs, tot_derivR, relatsR, phis, a2, bR, atau, btau2R, 3, draws, 0, 0,0);

act_draws = ceil(draws - 1000);
betasI = over_betasI(act_draws:end,:);
betasR = over_betasR(act_draws:end,:);
T_Build = toc(T_BuildS)

%Plots the fit of the model to the training data
GPplot2I = coverage(betasI,I_inputs,tot_derivI,phis,matrixI,draws/25,1); 
GPplot2R = coverage(betasR,R_inputs,tot_derivR,phis,matrixR,draws/25,1);
