%SIR Calculation of the Bounds Using RK4 for fixed time step. 
%Fifty random draws (curves) are generated, with the draws the furthest 
%above and below the mean prediction thrown out. Next highest and 
%lowest draws at each point are used as the bounds.

%Load the GP model for the SIR derivatives created in SIR_GP_Model
load('SIR GP Model.mat')
%Load the test curves
[curves,inputs, pureinputs]=SIR_GP_Data_Test(0);
%Load smoothing spline
x = dlmread('spline_coefficient_500.txt');
phis = splineconvert500(x);
%Load the training data for normalization
[~,~, pi]=SIR_GP_Data_Train(0);
pii = [];
for i=1:length(pi)
pii = [pii;pi{i}];
end

%Calculate the norms using the training data
norms = [0.5 min(pii(:,4)) min(pii(:,4)); 9 max(pii(:,4)) max(pii(:,5))];

%Calculate the mean model
mean_betasI = mean(betasI);
mean_betasR = mean(betasR);

%Generate 50 random integers between 1 and the number of draws to use in
%selecting which draws to calculate
a1 = randperm(size(betasI,1),50);
a2 = randperm(size(betasR,1),50);

%Initalize cells for the model predictions and bounds of each test curve
IR = cell(50,length(curves));
Imax = cell(length(curves),1);
Imin = cell(length(curves),1);
Rmax = cell(length(curves),1);
Rmin = cell(length(curves),1);

for i=1:length(curves)

    start = 0; %Start point of integration
    stop = inputs{i}(end,1); %End point of integration
    y0 = curves{i}(1,2:3)'; %Initial condition
    b = inputs{i}(:,2); %Vector of b values over the integration time span (the forcing function)
    h=(stop-start)/3500; %Step size

    test_I = inputs{i}(:,4); %I values
    test_R = inputs{i}(:,5); %R values
    test_t = inputs{i}(:,1); %time

    %Combine the models (mean and regular) for I and R into a pair of inputs for the RK4
    %integrator
    mean_betas = {mean_betasI;mean_betasR};
    matrix = {matrixI;matrixR};
    betas= {betasI;betasR};

    %Returns the Predictions / Integration of the mean model
    [Th1,mIR] = rk4_mean(mean_betas,matrix,b,norms,phis,start,stop,y0,h);

    %returns the Predictions / Integration of the 50 random models
    for j=1:50
        [Th1,IR{j,i}] = rk4(betas,matrix,b,norms,phis,start,stop,y0,h,a1(j),a2(j));
    end

    above_1=0;
    below_1=0;
    above_2=0;
    below_2=0;
    most_above_1 = 0;
    most_below_1 = 0;
    most_above_2 = 0;
    most_below_2 = 0;
    for k=1:50
        for l=1:length(IR{k,i})
            if (IR{k,i}(1,l)-mIR(1,l)) > above_1
                above_1 = (IR{k,i}(1,l)-mIR(1,l));
                most_above_1 = k;
            else
                if (IR{k,i}(1,l)-mIR(1,l)) < below_1
                    below_1 = (IR{k,i}(1,l)-mIR(1,l));
                    most_below_1 = k;
                end
            end
            if (IR{k,i}(2,l)-mIR(2,l)) > above_2
                above_2 = (IR{k,i}(2,l)-mIR(2,l));
                most_above_2 = k;
            else
                if (IR{k,i}(2,l)-mIR(2,l)) < below_2
                    below_2 = (IR{k,i}(2,l)-mIR(2,l));
                    most_below_2 = k;
                end
            end
        end
    end


    ind1 = 1;
    ind2 = 1;
    for j=1:length(IR)
        if j~= most_above_1 && j~= most_below_1
            Ib(ind1,:) = IR{j,i}(1,:);
            ind1 = ind1 + 1;
        end
        if j~= most_above_2 && j~= most_below_2
            Rb(ind2,:) = IR{j,i}(2,:);
            ind2 = ind2+1;
        end
    end

    Imax{i} = max(Ib);
    Imin{i} = min(Ib);
    Rmax{i} = max(Rb);
    Rmin{i} = min(Rb);


    %Produce plots of each test curve displaying the bounds, mean
    %prediction, and actual values of I and R for each test curve
    figure
    plot(Th1,mIR(1,:),'r')
    hold on
    plot(test_t,test_I,'bo','MarkerSize',2)
    hold on
    plot(Th1,Imax{i}, '-.', 'Color', [0.5 0.5 0.5])
    hold on
    plot(Th1,Imin{i},'-.', 'Color', [0.5 0.5 0.5])
    hold on
    legend('Mean Prediction','Actual Value','Bounds')
    xlabel('Time')
    ylabel('Population Infected')
    hold off

    figure
    plot(Th1,mIR(2,:),'r')
    hold on
    plot(test_t,test_R,'bo','MarkerSize',2)
    hold on
    plot(Th1,Rmax{i},'-.', 'Color', [0.5 0.5 0.5])
    hold on
    plot(Th1,Rmin{i},'-.', 'Color', [0.5 0.5 0.5])
    hold on
    legend('Mean Prediction','Actual Value','Bounds')
    xlabel('Time')
    ylabel('Population Recovered')
    hold off

end

    



function [T,Y]=rk4(betas,matrix,b,norms,phis,start,stop,y0,h,a1,a2)%Inputs are h1,h2,u
f=@(t,y,b_ind) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(b(b_ind),norms(1,1),norms(2,1))], betas{1}(a1,:), phis, matrix{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2))], betas{2}(a2,:), phis, matrix{2})];

T=start:h:stop;
y=y0;
Y=[y0,zeros(2,length(T)-1)];
ind = 2;
for t=T(1:end-1)
    if(t<h*10)
        b_ind = 1;
    else
        b_ind = floor(t/(h*10));
    end
    dy1=f(t,y,b_ind)*h; %euler estimate of y(t+h)-y(t)
    %In between each estimate of the derivative we check to see if the
    %value of the outputs exceeds the range of the training data, and if
    %the derivative would produce a new prediction further outside the
    %accepted range. If both conditions are true, the derivative is set
    %equal to 0.
    if y(1)>norms(2,2) && dy1(1)>0
        dy1(1)=0;
    else
        if y(1)<norms(1,2) && dy1(1)<0
            dy1(1)=0;
        end
    end
    if y(2)>norms(2,3) && dy1(2)>0
        dy1(2)=0;
    else
        if y(2)<norms(1,3) && dy1(2)<0
            dy1(2)=0;
        end
    end
    dy2=f(t+h/2,y+dy1/2,b_ind)*h; %midpoint estimate
    if (y(1)+dy1(1)/2)>norms(2,2) && dy2(1)>0
        dy2(1)=0;
    else
       if (y(1)+dy1(1)/2)<norms(1,2) && dy2(1)<0
            dy2(1)=0;
        end
    end
    if (y(2)+dy1(2)/2)>norms(2,3) && dy2(2)>0
        dy2(2)=0;
    else
        if (y(2)+dy1(2)/2)<norms(1,3) && dy2(2)<0
            dy2(2)=0;
        end
    end
    dy3=f(t+h/2,y+dy2/2,b_ind)*h; %refined midpoint method
    if (y(1)+dy2(1)/2)>norms(2,2) && dy3(1)>0
        dy3(1)=0;
    else
        if (y(1)+dy2(1)/2)<norms(1,2) && dy3(1)<0
            dy3(1)=0;
        end
    end
    if (y(2)+dy2(2)/2)>norms(2,3) && dy3(2)>0
        dy3(2)=0;
    else
        if (y(2)+dy2(2)/2)<norms(1,3) && dy3(2)<0
            dy3(2)=0;
        end
    end
    dy4=f(t+h,y+dy3,b_ind)*h; %forward estimate
    if (y(1)+dy3(1))>norms(2,2) && dy4(1)>0
        dy4(1)=0;
    else
        if (y(1)+dy3(1))<norms(1,2) && dy4(1)<0
            dy4(1)=0;
        end
    end
    if (y(2)+dy3(2))>norms(2,3) && dy4(2)>0
        dy4(2)=0;
    else
        if (y(2)+dy3(2))<norms(1,3) && dy4(2)<0
            dy4(2)=0;
        end
    end
    y=y+(dy1+2*dy2+2*dy3+dy4)/6; %averaging
    Y(:,ind)=y; %augment with a new column
    ind = ind+1;
end
end

function [T,Y]=rk4_mean(mean_betas,matrix,b,norms,phis,start,stop,y0,h)%Inputs is I,R,b

f=@(t,y,b_ind) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(b(b_ind),norms(1,1),norms(2,1))], mean_betas{1}, phis, matrix{1});...
             bss_eval([norma(y(1),norms(1,2),norms(2,2))], mean_betas{2}, phis, matrix{2})];

T=start:h:stop;
y=y0;
Y=[y0,zeros(2,length(T)-1)];
ind = 2;
for t=T(1:end-1)
    if(t<h*10)
        b_ind = 1;
    else
        b_ind = floor(t/(h*10));
    end
    dy1=f(t,y,b_ind)*h; %euler estimate of y(t+h)-y(t)
    if y(1)>norms(2,2) && dy1(1)>0
        dy1(1)=0;
    else
        if y(1)<norms(1,2) && dy1(1)<0
            dy1(1)=0;
        end
    end
    if y(2)>norms(2,3) && dy1(2)>0
        dy1(2)=0;
    else
        if y(2)<norms(1,3) && dy1(2)<0
            dy1(2)=0;
        end
    end
    dy2=f(t+h/2,y+dy1/2,b_ind)*h; %midpoint estimate
    if (y(1)+dy1(1)/2)>norms(2,2) && dy2(1)>0
        dy2(1)=0;
    else
       if (y(1)+dy1(1)/2)<norms(1,2) && dy2(1)<0
            dy2(1)=0;
        end
    end
    if (y(2)+dy1(2)/2)>norms(2,3) && dy2(2)>0
        dy2(2)=0;
    else
        if (y(2)+dy1(2)/2)<norms(1,3) && dy2(2)<0
            dy2(2)=0;
        end
    end
    dy3=f(t+h/2,y+dy2/2,b_ind)*h; %refined midpoint method
    if (y(1)+dy2(1)/2)>norms(2,2) && dy3(1)>0
        dy3(1)=0;
    else
        if (y(1)+dy2(1)/2)<norms(1,2) && dy3(1)<0
            dy3(1)=0;
        end
    end
    if (y(2)+dy2(2)/2)>norms(2,3) && dy3(2)>0
        dy3(2)=0;
    else
        if (y(2)+dy2(2)/2)<norms(1,3) && dy3(2)<0
            dy3(2)=0;
        end
    end
    dy4=f(t+h,y+dy3,b_ind)*h; %forward estimate
    if (y(1)+dy3(1))>norms(2,2) && dy4(1)>0
        dy4(1)=0;
    else
        if (y(1)+dy3(1))<norms(1,2) && dy4(1)<0
            dy4(1)=0;
        end
    end
    if (y(2)+dy3(2))>norms(2,3) && dy4(2)>0
        dy4(2)=0;
    else
        if (y(2)+dy3(2))<norms(1,3) && dy4(2)<0
            dy4(2)=0;
        end
    end
    y=y+(dy1+2*dy2+2*dy3+dy4)/6; %averaging
    Y(:,ind)=y; %augment with a new column
    ind = ind+1;
end
end

function ny = norma(y,n1,n2) %0 to 1 Normalization subfunction.
%y is the actual value, n1 is the minimum, n2 is the maximum
if y > n2
    ny= 1;
else
    if y < n1
        ny = 0;
    else
        ny = (y-n1)/(n2-n1);
    end
end
end