%SIR GP Model Integration Using RK4 for fixed time step
%Selected results from the test curves were saved and plotted in more
%detail in the script INSERT NAME HERE

%Load the model created of the derivatives in INSERT CODE NAME HERE 
load('SIR GP Model.mat')
[curves,inputs, pureinputs]=SIR_GP_Data_Test(0); %Load the test data
x = dlmread('spline_coefficient_500.txt');
phis = splineconvert500(x);
[~,~, pi]=SIR_GP_Data_Train(0); %Load the inputs of the training data for use in denormalization
pii = [];
for i=1:length(pi)
pii = [pii;pi{i}];
end

norms = [0.5 min(pii(:,4)) min(pii(:,4)); 9 max(pii(:,4)) max(pii(:,5))]; %Minimum and maximum values of the training data

%Calculate the mean GP model 
mean_betasI = mean(betasI); 
mean_betasR = mean(betasR);

%Initialize measurement values
MAE_I = zeros(length(curves),1);
MAE_R = zeros(length(curves),1);
MAPE_I = zeros(length(curves),1);
MAPE_R = zeros(length(curves),1);

T_Predict = zeros(length(curves),1);

%Make predictions for all curves in the test set and evaluate the error
for i=1:length(curves)
    %Initalize start and stop times for each test curve
    start = 0;
    stop = inputs{i}(end,1);
    %Pull inital condition from the 1st entry in the test curve
    y0 = curves{i}(1,2:3)';
    %b is the vector of the forcing function input
    b = inputs{i}(:,2);
    %Determine RK4 step size
    h=(stop-start)/3500;

    %Actual values of the test data to be used in comparison / calculation
    %of the error
    test_I = pureinputs{i}(:,4);
    test_R = pureinputs{i}(:,5);
    test_t = inputs{i}(:,1);

    %Combine the two models for I and R into one set of inputs to provide
    %to the RK4 function
    mean_betas = {mean_betasI;mean_betasR};
    matrix = {matrixI;matrixR};

    %Integrate the model (and measure the time it takes)
    T_PredictS = tic;
    [Th1,h1] = rk4(mean_betas,matrix,b,norms,phis,start,stop,y0,h);
    T_Predict(i) = toc(T_PredictS);


    %Plot the results of each curve against their actual values
    figure
    plot(Th1,h1(1,:) ,'b')
    hold on
    plot(test_t,test_I,'r')
    xlabel('Time (Weeks)')
    ylabel('Infected Individuals')
    legend('GP Mean Model','Actual')
    hold off
    figure
    plot(Th1,h1(2,:),'b')
    hold on
    plot(test_t,test_R,'r')
    xlabel('Time (Weeks)')
    ylabel('Recovered Individuals')
    legend('GP Mean Model','Actual')
    hold off

    %Reduce the number of datapoints from the integrated model so as to be
    %comparible to the true values
    I_comp = h1(1,1:10:end)';
    R_comp = h1(2,1:10:end)';

    %Calculate the MAE, MAPEs for each curve in the test set
    MAE_I(i) = mean(abs(I_comp-test_I));
    MAPE_I(i) = mean(abs((I_comp-test_I)./test_I));
    MAE_R(i) = mean(abs(R_comp-test_R));
    MAPE_R(i) = mean(abs((R_comp-test_R)./test_R));

end

%Calculate the mean values of the MAEs and MAPEs for all curves
MAE_I_tot = mean(MAE_I)
MAE_R_tot = mean(MAE_R)
MAPE_I_tot = mean(MAPE_I)
MAPE_R_tot = mean(MAPE_R)

%Calculate the standard deviation of the MAEs and MAPEs between curves in
%the test set
MAE_R_stdev = (sum((MAE_R-MAE_R_tot).^2)/24)^0.5
MAE_I_stdev = (sum((MAE_I-MAE_I_tot).^2)/24)^0.5
MAPE_R_stdev = (sum((MAPE_R-MAPE_R_tot).^2)/24)^0.5
MAPE_I_stdev = (sum((MAPE_I-MAPE_I_tot).^2)/24)^0.5

%Calculate and display the average time it takes to produce the integrated
%model's predictions
Mean_Predict_Time = mean(T_Predict)


function [T,Y]=rk4(mean_betas,matrix,b,norms,phis,start,stop,y0,h)%Inputs are I,R,b

f=@(t,y,b_ind) [bss_eval([norma(y(1),norms(1,2),norms(2,2)),norma(y(2),norms(1,3),norms(2,3)),norma(b(b_ind),norms(1,1),norms(2,1))], mean_betas{1}, phis, matrix{1});...
    bss_eval([norma(y(1),norms(1,2),norms(2,2))], mean_betas{2}, phis, matrix{2})];

T=start:h:stop;
y=y0;
Y=[y0,zeros(2,length(T)-1)];
ind = 2;
for t=T(1:end-1)
    if(t<h*10)
        b_ind = 1;
    else
        b_ind = floor(t/(h*10));
    end
    dy1=f(t,y,b_ind)*h; %euler estimate of y(t+h)-y(t)
    %In between each estimate of the derivative we check to see if the
    %value of the outputs exceeds the range of the training data, and if
    %the derivative would produce a new prediction further outside the
    %accepted range. If both conditions are true, the derivative is set
    %equal to 0.
    if y(1)>norms(2,2) && dy1(1)>0
        dy1(1)=0;
    else
        if y(1)<norms(1,2) && dy1(1)<0
            dy1(1)=0;
        end
    end
    if y(2)>norms(2,3) && dy1(2)>0
        dy1(2)=0;
    else
        if y(2)<norms(1,3) && dy1(2)<0
            dy1(2)=0;
        end
    end
    dy2=f(t+h/2,y+dy1/2,b_ind)*h; %midpoint estimate
    if (y(1)+dy1(1)/2)>norms(2,2) && dy2(1)>0
        dy2(1)=0;
    else
        if (y(1)+dy1(1)/2)<norms(1,2) && dy2(1)<0
            dy2(1)=0;
        end
    end
    if (y(2)+dy1(2)/2)>norms(2,3) && dy2(2)>0
        dy2(2)=0;
    else
        if (y(2)+dy1(2)/2)<norms(1,3) && dy2(2)<0
            dy2(2)=0;
        end
    end
    dy3=f(t+h/2,y+dy2/2,b_ind)*h; %refined midpoint method
    if (y(1)+dy2(1)/2)>norms(2,2) && dy3(1)>0
        dy3(1)=0;
    else
        if (y(1)+dy2(1)/2)<norms(1,2) && dy3(1)<0
            dy3(1)=0;
        end
    end
    if (y(2)+dy2(2)/2)>norms(2,3) && dy3(2)>0
        dy3(2)=0;
    else
        if (y(2)+dy2(2)/2)<norms(1,3) && dy3(2)<0
            dy3(2)=0;
        end
    end
    dy4=f(t+h,y+dy3,b_ind)*h; %forward estimate
    if (y(1)+dy3(1))>norms(2,2) && dy4(1)>0
        dy4(1)=0;
    else
        if (y(1)+dy3(1))<norms(1,2) && dy4(1)<0
            dy4(1)=0;
        end
    end
    if (y(2)+dy3(2))>norms(2,3) && dy4(2)>0
        dy4(2)=0;
    else
        if (y(2)+dy3(2))<norms(1,3) && dy4(2)<0
            dy4(2)=0;
        end
    end
    y=y+(dy1+2*dy2+2*dy3+dy4)/6; %averaging
    Y(:,ind)=y; %augment with a new column
    ind = ind+1;
end
end

function ny = norma(y,n1,n2) %0 to 1 Normalization subfunction.
%y is the actual value, n1 is the minimum, n2 is the maximum
if y > n2
    ny= 1;
else
    if y < n1
        ny = 0;
    else
        ny = (y-n1)/(n2-n1);
    end
end
end